class DuplicateExternalLinkException(Exception):
    def __init__(self, message, external_id=None):
        super().__init__(message)
        self.external_id = external_id


class DuplicateFileException(Exception):
    def __init__(self, message, filename=None, md5sum=None):
        super().__init__(message)
        self.message = message
        self.filename = filename
        self.md5sum = md5sum


class ColumnNotFoundException(Exception):
    pass


class FileNotFoundException(Exception):
    pass


class IncorrectTypeException(Exception):
    pass


class InvalidExtensionException(Exception):
    pass


class InvalidObjectException(Exception):
    pass


class NoMediafilesException(Exception):
    pass


class NonUniqueException(Exception):
    pass


class NotFoundException(Exception):
    pass


class NoTenantException(Exception):
    pass


class UnsupportedVersionException(Exception):
    pass



from configuration import get_collection_mapper
from flask_restful import abort
from inuits_policy_based_auth import BaseAuthorizationPolicy, RequestContext
from inuits_policy_based_auth.contexts import UserContext, PolicyContext
from storage.storagemanager import StorageManager


class MultiTenantPolicy(BaseAuthorizationPolicy):
    """
    An authorization policy that checks if you have access to the requested
    item or operation through your tenant.
    """

    def authorize(
        self,
        policy_context: PolicyContext,
        user_context: UserContext,
        request_context: RequestContext,
    ) -> PolicyContext:
        """
        Authorizes a user.

        Parameters
        ----------
        policy_context : PolicyContext
            An object containing data about the context of an applied
            authorization policy.
        user_context : UserContext
            An object containing data about the authenticated user.
        request_context : RequestContext
            An object containing data about the context of a request.

        Returns
        -------
        PolicyContext
            An object containing data about the context of an applied
            authorization policy.
        """
        request = request_context.http_request
        view_args = request.view_args or {}
        item_id = view_args.get("id")
        policy_context.access_verdict = True
        if item_id:
            storage = StorageManager().get_db_engine()
            request_name = request.path.split("/")[1]
            collection = get_collection_mapper().get(request_name, request_name)
            item = storage.get_item_from_collection_by_id(collection, item_id)
            if not item:
                abort(
                    404,
                    message=f"Item with id {id} doesn't exist in collection {collection}",
                )
            item_relations = storage.get_collection_item_relations(collection, item_id)
            if (
                item.get("type") != "ticket"
                and not any(
                    x
                    for x in item_relations
                    if x["type"] == "isIn"
                    and x["key"] == user_context.x_tenant.raw["_id"]
                )
                and collection != "mediafiles"
            ):
                policy_context.access_verdict = False
        if "/filter" in request.path:
            user_context.access_restrictions.filters = [
                {
                    "type": "selection",
                    "parent_key": "relations",
                    "key": "isIn",
                    "value": [user_context.x_tenant.raw["_id"]],
                    "match_exact": True,
                }
            ]
        else:
            user_context.access_restrictions.filters = {
                "relations.type": "isIn",
                "relations.key": user_context.x_tenant.raw["_id"],
            }
        return policy_context



from abc import ABC, abstractmethod
from copy import deepcopy
from elody.migration.base_object_migrator import BaseObjectMigrator


class BaseObjectConfiguration(ABC):
    SCHEMA_TYPE = "elody"
    SCHEMA_VERSION = 1

    @abstractmethod
    def crud(self):
        return {
            "collection": "entities",
            "collection_history": "history",
            "creator": lambda post_body, **kwargs: post_body,
            "nested_matcher_builder": lambda object_lists, keys_info, value: self.__build_nested_matcher(
                object_lists, keys_info, value
            ),
            "post_crud_hook": lambda **kwargs: None,
            "pre_crud_hook": lambda **kwargs: None,
            "storage_type": "db",
        }

    @abstractmethod
    def document_info(self):
        return {"object_lists": {"metadata": "key", "relations": "type"}}

    @abstractmethod
    def logging(self, flat_document, **kwargs):
        info_labels = {
            "uuid": flat_document.get("_id"),
            "type": flat_document.get("type"),
            "schema": f"{flat_document.get('schema.type')}:{flat_document.get('schema.version')}",
        }
        try:
            user_context = kwargs.get("get_user_context")()  # pyright: ignore
            info_labels["http_method"] = user_context.bag.get("http_method")
            info_labels["requested_endpoint"] = user_context.bag.get(
                "requested_endpoint"
            )
            info_labels["full_path"] = user_context.bag.get("full_path")
            info_labels["preferred_username"] = user_context.preferred_username
            info_labels["email"] = user_context.email
            info_labels["user_roles"] = ", ".join(user_context.x_tenant.roles)
            info_labels["x_tenant"] = user_context.x_tenant.id
        except Exception:
            pass
        return {"info_labels": info_labels, "loki_indexed_info_labels": {}}

    @abstractmethod
    def migration(self):
        return BaseObjectMigrator(status="disabled")

    @abstractmethod
    def serialization(self, from_format, to_format):
        def serializer(document, **_):
            return document

        return serializer

    @abstractmethod
    def validation(self):
        def validator(http_method, content, **_):
            pass

        return "function", validator

    def _get_merged_post_body(self, post_body, document_defaults, object_list_name):
        key = self.document_info()["object_lists"][object_list_name]
        post_body[object_list_name] = self.__merge_object_lists(
            document_defaults.get(object_list_name, []),
            post_body.get(object_list_name, []),
            key,
        )
        return post_body

    def _sanitize_document(self, document, object_list_name, value_field_name):
        object_list = deepcopy(document[object_list_name])
        for element in object_list:
            if not element[value_field_name]:
                document[object_list_name].remove(element)

    def _sort_document_keys(self, document):
        def sort_keys(data):
            if isinstance(data, dict):
                sorted_items = {key: data.pop(key) for key in sorted(data.keys())}
                for key, value in sorted_items.items():
                    data[key] = sort_keys(value)
                return data
            elif isinstance(data, list):
                if all(isinstance(i, str) for i in data):
                    data.sort()
                    return data
                else:
                    for index, item in enumerate(data):
                        data[index] = sort_keys(item)
                    return data
            else:
                return data

        for key, value in self.document_info()["object_lists"].items():
            document[key] = sorted(document[key], key=lambda property: property[value])
        sort_keys(document)

    def __build_nested_matcher(self, object_lists, keys_info, value, index=0):
        if index == 0 and not any(info["is_object_list"] for info in keys_info):
            if value in ["ANY_MATCH", "NONE_MATCH"]:
                value = {"$exists": value == "ANY_MATCH"}
            return {".".join(info["key"] for info in keys_info): value}

        info = keys_info[index]

        if info["is_object_list"]:
            nested_matcher = self.__build_nested_matcher(
                object_lists, keys_info, value, index + 1
            )
            elem_match = {
                "$elemMatch": {
                    object_lists[info["key"]]: info["object_key"],
                    keys_info[index + 1]["key"]: nested_matcher,
                }
            }
            if value in ["ANY_MATCH", "NONE_MATCH"]:
                del elem_match["$elemMatch"][keys_info[index + 1]["key"]]
                if value == "NONE_MATCH":
                    return {"NOR_MATCHER": {info["key"]: {"$all": [elem_match]}}}
            return elem_match if index > 0 else {info["key"]: {"$all": [elem_match]}}

        return value

    def __merge_object_lists(self, source, target, key):
        for target_item in target:
            for source_item in source:
                if source_item[key] == target_item[key]:
                    source.remove(source_item)
        return [*source, *target]
